"""Rougail's tools

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2023

distribued with GPL-2 or later license

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
from typing import List, Union
from unicodedata import normalize, combining
import re

from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader, module_from_spec

from jinja2 import DictLoader, TemplateSyntaxError
from jinja2.sandbox import SandboxedEnvironment
from jinja2.parser import Parser
from jinja2.nodes import Getattr

from .i18n import _
from .error import DictConsistencyError

NAME_REGEXP = re.compile(r"^[a-z0-9_]*$")


def valid_variable_family_name(name: str,
                               xmlfiles: List[str],
                               ) -> None:
    match = NAME_REGEXP.search(name)
    if not match:
        msg = _(f'invalid variable or family name "{name}" must only contains '
                'lowercase ascii character, number or _')
        raise DictConsistencyError(msg, 76, xmlfiles)


def normalize_family(family_name: str) -> str:
    """replace space, accent, uppercase, ... by valid character
    """
    if not family_name:
        return
    family_name = family_name.replace('-', '_').replace(' ', '_').replace('.', '_')
    nfkd_form = normalize('NFKD', family_name)
    family_name = ''.join([c for c in nfkd_form if not combining(c)])
    return family_name.lower()


def load_modules(eosfunc_file) -> List[str]:
    """list all functions in eosfunc
    """
    loader = SourceFileLoader('eosfunc', eosfunc_file)
    spec = spec_from_loader(loader.name, loader)
    eosfunc = module_from_spec(spec)
    loader.exec_module(eosfunc)
    return eosfunc


def get_realpath(path: str,
                 path_prefix: str,
                 ) -> str:
    if path_prefix:
        return f'{path_prefix}.{path}'
    return path


def get_jinja_variable_to_param(jinja_text,
                                objectspace,
                                xmlfiles,
                                functions,
                                path_prefix,
                                ):
    try:
        env = SandboxedEnvironment(loader=DictLoader({'tmpl': jinja_text}))
        env.filters = functions
        parsed_content = Parser(env, jinja_text, "", "").parse()
        def recurse_getattr(g: Getattr):
            if isinstance(g.node, Getattr):
                return recurse_getattr(g.node) + "." + g.attr
            return g.node.name + "." + g.attr
        variables = set()
        for g in parsed_content.find_all(Getattr):
            variables.add(recurse_getattr(g))
    except TemplateSyntaxError as err:
        msg = _(f'error in jinja "{jinja_text}": {err}')
        raise Exception(msg) from err
    variables = list(variables)
    variables.sort()
    for variable_path in variables:
        variable, suffix = objectspace.paths.get_with_dynamic(get_realpath(variable_path, path_prefix))
        if variable and variable.path in objectspace.variables:
            yield variable, suffix, variable_path


class CalculationOld:
    def __init__(self,
                 object_name: str,
                 attribute: str,
                 path: str,
                 datas: Union[dict, List[dict]],
                 filenames: List[str],
                 objectspace: 'RougailConvert',
                 path_prefix: str,
                 inside_list: bool,
                 ) -> None:
        self.path_prefix = path_prefix
        if 'type' not in datas:
            raise Exception(f'the {object_name} "{path}" in "{filenames}" must have a "type" attribut for "{attribute}"')
        self.type = datas.pop('type')
        self.path = path
        self.attribute = attribute
        if self.type != 'suffix':
            self.value = datas.pop(self.type)
        if 'params' in datas:
            if self.type == 'variable':
                raise Exception('param not allowed with type variable')
            self.params = datas.pop('params')
        else:
            self.params = {}
        if 'return_type' in datas:
            self.return_type = datas.pop('return_type')
            if self.return_type not in objectspace.variable_types:
                raise Exception(f'unknown "return_type" in {attribute} of variable "{self.path}"')
        else:
            self.return_type = None
        if self.type == 'variable':
            self.propertyerror = datas.pop('propertyerror', True)
        if datas:
            raise Exception('hu.?')
        self.objectspace = objectspace
        self.inside_list = inside_list

    def get_realpath(self,
                     path: str,
                     ) -> str:
        return get_realpath(path, self.path_prefix)

    def change_variable_in_params(self):
        if self.params:
            delete_keys = []
            for key, value in self.params.items():
                if isinstance(value, dict):
                    if value['type'] == 'variable':
                        variable_path = self.get_realpath(value['variable'])
                        variable, suffix = self.objectspace.paths.get_with_dynamic(variable_path)
                        if variable:
                            if not isinstance(variable, self.objectspace.variable):
                                print("pfff it's a family")
                            self.params[key]['variable'] = variable
                            if suffix:
                                self.params[key]['suffix'] = suffix
                        else:
                            if not value.get('optional'):
                                raise Exception(f'pffff {value["variable"]}')
                            delete_keys.append(key)
                    elif value['type'] == 'information':
                        if 'variable' in value:
                            variable_path = self.get_realpath(value['variable'])
                            value['variable'] = self.objectspace.paths[variable_path]
            for key in delete_keys:
                del self.params[key]


    def _jinja_to_function(self,
                           function,
                           return_type,
                           multi,
                           *,
                           add_help=False,
                           ):
        variable = self.objectspace.paths[self.path]
        jinja_path = f'{self.attribute}_{self.path}'
        idx = 0
        while jinja_path in self.objectspace.jinja:
            jinja_path = f'{self.attribute}_{self.path}_{idx}'
            idx += 1
        self.objectspace.jinja[jinja_path] = self.value
        default = {'function': function,
                   'params': {'__internal_jinja': jinja_path,
                              '__internal_type': return_type,
                              '__internal_multi': multi,
                              }
                   }
        if add_help:
            default['help'] = function + '_help'
        if self.params:
            default['params'] |= self.params
        for sub_variable, suffix, true_path in get_jinja_variable_to_param(self.value,
                                                                           self.objectspace,
                                                                           variable.xmlfiles,
                                                                           self.objectspace.functions,
                                                                           self.path_prefix,
                                                                           ):
            if isinstance(sub_variable, self.objectspace.variable):
                default['params'][true_path] = {'type': 'variable',
                                                'variable': sub_variable,
                                                }
                if suffix:
                    default['params'][true_path]['suffix'] = suffix
        return default

    def to_property_function(self) -> dict:
        self.change_variable_in_params()
        if self.return_type:
            raise Exception('return_type not allowed!')
        if self.type == 'jinja':
            if None in self.params:
                self.params[None] = [self.attribute] + self.params[None]
            else:
                self.params[None] = [self.attribute]
            return self._jinja_to_function('jinja_to_property',
                                           'string',
                                           False,
                                           add_help=True,
                                           )
        if self.type == 'variable':
            variable_path = self.get_realpath(self.value)
            variable = self.objectspace.paths[variable_path]
            if variable.type != 'boolean':
                raise Exception('only boolean!')
            print("**********", self.attribute)
            default = {'function': 'variable_to_property',
                       'params': {None: [self.attribute, variable]},
#                    'params': {None: [{'type': 'variable',
#                                      'variable': self.objectspace.paths[variable_path],
#                                      'propertyerror': self.propertyerror,
#                                      }],
                       'help': 'variable_to_property',
                       }
            print('pfff', default, self.path)
            return default
        raise Exception('pfff')

    def to_default_function(self) -> dict:
        self.change_variable_in_params()
        if self.type == 'jinja':
            if self.return_type:
                return_type = self.return_type
            else:
                variable = self.objectspace.paths[self.path]
                return_type = variable.type
            if self.inside_list:
                multi = False
            elif self.path in self.objectspace.followers:
                multi = self.objectspace.multis[self.path] == 'submulti'
            else:
                multi = self.path in self.objectspace.multis
            return self._jinja_to_function('jinja_to_function',
                                           return_type,
                                           multi,
                                           )
        elif self.type == 'variable':
            variable_path = self.get_realpath(self.value)
            return {'function': 'calc_value',
                    'params': {None: [{'type': 'variable',
                                      'variable': self.objectspace.paths[variable_path],
                                      'propertyerror': self.propertyerror,
                                      }],
                               'multi': self.path in self.objectspace.multis,
                               'allow_none': True,
                               }
                       }
        elif self.type == 'suffix':
            if self.params:
                raise Exception('pfff')
            variable_path = self.get_realpath(self.value)
            return {'function': 'calc_value',
                    'params': {None: [{'type': 'suffix'}],
                               'multi': self.objectspace.paths[variable_path].multi,
                               }
                    }
        else:
            raise Exception('pfff')

    def to_choice_function(self) -> dict:
        self.change_variable_in_params()
        return_type = self.return_type
        if return_type is None:
            return_type = 'string'
        if self.type == 'jinja':
            return self._jinja_to_function('jinja_to_function',
                                           return_type,
                                           not self.inside_list,
                                           )
        elif self.type == 'variable':
            variable_path = self.get_realpath(self.value)
            variable = self.objectspace.paths[variable_path]
            if variable.path not in self.objectspace.multis:
                raise Exception('pffff')
            return {'function': 'calc_value',
                    'params': {None: [{'type': 'variable',
                                       'variable': variable,
                                       }],
                               'multi': False,
                               'allow_none': True,
                               }
                       }
        else:
            raise Exception('pfff')
